/*
Jos Angel Estvez Armas
Prog. Paralelo II
"TSP"
Version Secuencial-Iterativo
pila
*/

/*#include <curses.h>*/
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#define MAX_NODES 50
#define MAX_VALUE 500
#define INFINITO	-99
#define GRANDE		20000

typedef int tablas[MAX_NODES][MAX_NODES];
typedef int array[MAX_NODES];
typedef struct structPila {
	tablas A;
	array row, col, forwP, backP;
	int cost, edges;
	struct structPila * next;
	} pila;

/*------------- Variables globales --------------------*/

int best_solution, size, MOSTRAR;
array bestP;
pila * top;
int nump_res, nump_gen;

/*---------- Prototipos de funciones ------------------*/

int	leeTabla (char *f_name, tablas tabla);
void	camino();
void	muestra (tablas A, int n, array row, array col);
void	init();
void	copyMatrix(tablas A, tablas B);
void	copyArraySin(array A, array B, int exc, int n);
void	copyArray(array A, array B);
int	reduce(tablas A, int n, array row, array col);
int	bestEdge(tablas A, int n, int *r, int *c, array row, array col);
void	explore();
pila	* push();
pila	* pop();
int	emptyStack();
int	intercambio();

/*---------------------------------*/

main(argc, argv)
	int argc;
	char **argv;
	{
	array row, col, forwP, backP;
	tablas A;
	struct timeval time1[1], time2[1];
	double t1, t2;

	printf("\n");
	if (argc < 2)
		{
		printf("\nUsage Mode: %s file.dat\n", argv[0]);
		return 0;
		}
	if (! leeTabla(argv[1], A))
		{
		printf("\nError... File %s doesn't exists\n", argv[1]);
		return 0;
		}

	printf("\n:::::::: TSP - Sequential-Iterative ::::::::");
	if (argc > 2)
		MOSTRAR = 1;
	else
		MOSTRAR = 0;

	init(row, col, forwP, backP);
	if (MOSTRAR)
		{
		printf("\nMatrix:\n");
		muestra(A, size, row, col);
		}

	gettimeofday(time1, NULL);
	explore(A, row, col, forwP, backP);
	gettimeofday(time2, NULL);

	t1 = time2->tv_sec - time1->tv_sec;
 	t2 = t1 + (time2->tv_usec - time1->tv_usec) / 1.0e6;
	/*t2 = t2 / CLOCKS_PER_SEC;*/

	printf("\nThe best solution is %d.", best_solution);
	camino();
	printf("\nNum. problems gen.: %4d\t\tsolv.: %4d", nump_gen, nump_res);
	printf("\nGlobal time: %12.8f segs.\n", t2);

	return 1;
	}  /* main */

/*------------------------------------------------------------------------*/

int leeTabla (char *f_name, tablas tabla)
	{
	FILE *fd;
	int i, j;

	fd = fopen(f_name, "r");
	if (fd == NULL)
		return 0;
	fscanf(fd, "%d", &size);
	printf("\nPROBLEM SIZE : %d.", size);
	/* Read Weight */
	for (i = 0; i < size; i++)
		for (j = 0; j < size; j++)
			if (i != j)
				fscanf(fd, "%d", &tabla[i][j]);
			else
				tabla[i][j] = INFINITO;
	fclose(fd);

	return size;
	}

/*------------------------------------------------------------------------*/

void camino()
	{
	int i = 0;

	printf("\nRoute: %d", i+1);
	while ((INFINITO != bestP[i]) && (bestP[i] != 0))
		{
		i = bestP[i];
		printf(" - %d", i+1);
		}
	printf(" - %d\n", bestP[i]+1);
	}

/*------------------------------------------------------------------------*/

void init(row, col, forwP, backP)
	array row, col, forwP, backP;
	{
	int i;

	for (i = 0; i < size; i++)
		{
		row[i] = col[i] = i;
		backP[i] = forwP[i] = bestP[i] = INFINITO;
		}
	best_solution = GRANDE;
	top = NULL;
	nump_res = nump_gen = 0;
	}

/*------------------------------------------------------------------------*/

void copyMatrix(A, B)
	tablas A, B;
	{
	int i, j;

	for (i = 0; i < size; i++)
		for (j = 0; j < size; j++)
			B[i][j] = A[i][j];
	}

/*------------------------------------------------------------------------*/

void copyArraySin(A, B, exc, n)
	array A, B;
	int exc, n;
	{
	int i;

	for (i = 0; i < exc; i++)
		B[i] = A[i];
	for (i = exc; i < n-1; i++)
		B[i] = A[i+1];
	}

/*------------------------------------------------------------------------*/

void copyArray(A, B)
	array A, B;
	{
	int i;

	for (i = 0; i < size; i++)
		B[i] = A[i];
	}

/*------------------------------------------------------------------------*/

int reduce(tablas A, int n, array row, array col)
	{
	int i, j, min, rvalue = 0;

	/* filas */
	for (i = 0; i < n; i++)
		{
		/* primero se busca el mnimo elemento */
		min = GRANDE;
		for (j = 0; j < n; j++)
			if ((INFINITO != A[row[i]][col[j]]) && (A[row[i]][col[j]] < min))
				min = A[row[i]][col[j]];
		/* restamos el mnimo y aumentamos rvalue */
		if (min > 0)
			{
			for (j = 0; j < n; j++)
				if (A[row[i]][col[j]] != INFINITO)
					A[row[i]][col[j]] -= min;
			rvalue += min;
			}
		}
	/* columnas */
	for (j = 0; j < n; j++)
		{
		/* primero se busca el mnimo elemento */
		min = GRANDE;
		for (i = 0; i < n; i++)
			if ((INFINITO != A[row[i]][col[j]]) && (A[row[i]][col[j]] < min))
				min = A[row[i]][col[j]];
		/* restamos el mnimo y aumentamos rvalue */
		if (min > 0)
			{
			for (i = 0; i < n; i++)
				if (A[row[i]][col[j]] != INFINITO)
					A[row[i]][col[j]] -= min;
			rvalue += min;
			}
		}

	return rvalue;
	}

/*------------------------------------------------------------------------*/

int bestEdge(tablas A, int n, int *r, int *c, array row, array col)
	{
	int i, j, k, most = -GRANDE, minr, minc, total;

	for (i = 0; i < n; i++)
		for (j = 0; j < n; j++)
			if (! A[row[i]][col[j]])
				{
				/* buscar el mnimo de esa columna y fila (salvo esta) */
				minr = minc = GRANDE;
				for (k = 0; k < n; k++)
					{
					if ((INFINITO != A[row[i]][col[k]]) &&
									(A[row[i]][col[k]] < minr) && (col[k] != col[j]))
						minr = A[row[i]][col[k]];
					if ((INFINITO != A[row[k]][col[j]]) &&
									(A[row[k]][col[j]] < minc) && (row[k] != row[i]))
						minc = A[row[k]][col[j]];
					}
				total = minr + minc;
				if (total > most)
					{
					most = total;
					*r = i;
					*c = j;
					}
				}

	if (MOSTRAR)
		printf("\n (%d, %d). Cost: %d.", row[*r]+1, col[*c]+1, most);
	return most;
	}

/*------------------------------------------------------------------------*/

void muestra (tablas A, int n, array row, array col)
	{
	int i, j;

	printf("    ");
	for (j = 0; j < n; j++)
		printf("  <%2d>", col[j]+1);
	for (i = 0; i < n; i++)
		{
		printf("\n<%2d>", row[i]+1);
		for (j = 0; j < n; j++)
			if (A[row[i]][col[j]] == INFINITO)
				printf("   INF");
			else
				printf("   %3d", A[row[i]][col[j]]);
		}
	printf("\n");
	}

/*-----------------------------------------------------------*/

void explore(A, row, col, forwP, backP)
	tablas A;
	array row, col, forwP, backP;
	{
	int n, i, j, most, lowerbound, last, first;
	int edges, cost;
	pila * p;

	n = size;
	p = push();		/* creamos primer problema */
	copyMatrix(A, p->A);
	copyArray(row, p->row);
	copyArray(col, p->col);
	copyArray(forwP, p->forwP);
	copyArray(backP, p->backP);
	p->edges = 0;
	p->cost = 0;
	while (! emptyStack())
		{
		/* sacar un problema de la pila */
		copyMatrix(top->A, A);
		copyArray(top->row, row);
		copyArray(top->col, col);
		copyArray(top->forwP, forwP);
		copyArray(top->backP, backP);
		edges = top->edges;
		cost = top->cost;
		pop();
		n = size - edges;

		if (MOSTRAR)
			{
			printf("\nMatrix %dx%d...", n, n);
			printf("\n...before reduction...\n");
			muestra(A, n, row, col);
			}

		cost += reduce(A, n, row, col);

		if (MOSTRAR)
			{
			printf("\n...and after reduction.\n");
			muestra(A, n, row, col);
			printf("Cost: %d.", cost);
			}

		if (cost < best_solution)	/* hay una mejora */
			if (n == 2)			/* no se debe extraer ms */
				{
				for (i = 0; i < size; i++)
					bestP[i] = forwP[i];
				if (A[row[0]][col[0]] == INFINITO)
					{
					bestP[row[0]] = col[1];
					bestP[row[1]] = col[0];
					}
				else
					{
					bestP[row[0]] = col[0];
					bestP[row[1]] = col[1];
					}
				nump_res++;
				best_solution = cost;
				}
			else			/* se sigue explorando... */
				{
				most = bestEdge(A, n, &i, &j, row, col);
				lowerbound = cost + most;
				/* se mete el problema izquierdo en la pila */
				p = push();
				copyMatrix(A, p->A);
				copyArray(forwP, p->forwP);
				copyArray(backP, p->backP);
				p->forwP[row[i]] = col[j];
				p->backP[col[j]] = row[i];
				/* prevencin de ciclos */
				last = col[j];
				while (p->forwP[last] != INFINITO)
					last = p->forwP[last];
				first = row[i];
				while (p->backP[first] != INFINITO)
					first = p->backP[first];
				p->A[last][first] = INFINITO;
				/* eliminar columna y fila */
				copyArraySin(row, p->row, i, n);		/* copia el array saltandose i */
				copyArraySin(col, p->col, j, n);		/* copia el array saltandose j */
				p->cost = cost;
				p->edges = edges+1;
				nump_gen++;
				/* ya se insert el problema en la pila */
				if (lowerbound < best_solution)	/* en iterativo, casi siempre */
					{
					/* se mete problema derecho en la pila */
					p = push();
					copyMatrix(A, p->A);
					copyArray(forwP, p->forwP);
					copyArray(backP, p->backP);
					copyArray(row, p->row);
					copyArray(col, p->col);
					p->cost = cost;
					p->edges = edges;
					p->A[row[i]][col[j]] = INFINITO;
					nump_gen++;
					/* ya se insert el problema en la pila */
					intercambio();		/* se intercambia el TOP y NEXT */
					}
				}	/* fin del else */
		}	/* end of while */
	}	/* end of explore */

/*========================= FUNCIONES DE PILAS ===========================*/

pila * push()
	{
	pila * p;

	p = malloc(sizeof(pila));
	if (p == NULL)
		{
		printf("\nOut of memory\n");
		exit(0);
		}
	p->next = top;
	top = p;

	return p;
	}

/*------------------------------------------------------------------------*/

pila * pop()
	{
	pila * p;

	if (emptyStack())
		{
		printf("\n\aEmpty Stack!!!!!!!!!!!!!\n");
		return NULL;
		}

	p = top;
	top = top->next;
	free(p);

	return top;
	}

/*------------------------------------------------------------------------*/

int emptyStack()
	{
	if (top == NULL)
		return 1;

	return 0;
	}

/*------------------------------------------------------------------------*/

int intercambio()
	{
	pila * p;

	if (emptyStack())
		return 0;
	else
		if (top->next == NULL)
			return 0;

	/* todo va bien por aqu */
	p = top;
	top = top->next;
	p->next = top->next;
	top->next = p;
	return 1;
	}
